/*
 * Created on 10 Jul 2006
 * Created by Paul Gardner
 * Copyright (C) Azureus Software, Inc, All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

package com.biglybt.core.nat;

import java.net.InetSocketAddress;
import java.util.Map;

public interface
NATTraversalObserver
{
	public static final int	FT_NO_RENDEZVOUS	= 1;
	public static final int	FT_QUEUE_FULL		= 2;
	public static final int	FT_CANCELLED		= 3;

	public static String[] FT_STRINGS = {
		"Unknown", "No rendezvous", "Queue full", "Operation cancelled" };

	public void
	succeeded(
		InetSocketAddress	rendezvous,
		InetSocketAddress	target,
		Map					reply );

	public void
	failed(
		int			failure_type );

	public void
	failed(
		Throwable 	cause );

	public void
	disabled();
}
